/* Settings
 *
 * Authors: Florian Boor <fb@kernelconcepts.de>
 *          Gang Wang <gang.wang@orange-ftgroup.com>
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */

#include <../config.h>
                   
#include <glib.h>
#include <gconf/gconf-client.h>
#include <stdlib.h>

#include "settings.h"

#define MAX_CNXN 16

static GHashTable *hash_gconfclient = NULL; //mapping between setting and gconfclient
static GHashTable *hash_cnid_notify = NULL;  //mapping between connection id and SettingNotifyFunc
static GHashTable *hash_cnid_setting = NULL;//mapping between connection id and setting
static int setting = 0;
int cnid_array [MAX_CNXN];
static int cnid_array_cursor = 0;

static int cnid_array_lookup_cursor (int cnid)
{
  int i;
  for (i = 0; i < MAX_CNXN; i++)
  {
  	if (cnid_array [i] == cnid)
  	  return i;
  }
  return -1;
}

/*
 * Notify Object
 */
typedef struct _Notify Notify;

struct _Notify {
  SettingNotifyFunc func;
  gpointer data;
  GFreeFunc destroy_notify;
};

static Notify* notify_new(SettingNotifyFunc func,
                              GFreeFunc destroy_notify,
                              gpointer data);

static void notify_destroy(Notify* notify);

static Notify*
notify_new(SettingNotifyFunc func,
             GFreeFunc destroy_notify,
             gpointer data)
{
  Notify* notify;

  notify = g_new(Notify, 1);

  notify->func = func;
  notify->data = data;
  notify->destroy_notify = destroy_notify;
  
  return notify;
}

static void
notify_destroy(Notify* notify)
{
  g_return_if_fail(notify != NULL);

  if (notify->destroy_notify)
    (* notify->destroy_notify) (notify->data);

  g_free(notify);
}

/*
 *  Mapping functions
 */
static error_t error_mapping (GError *error)
{
  error_t err;
  
  if (error == NULL)
    return SETTING_ERROR_SUCCESS;
  
  gint gconf_error = error->code; 
  switch (gconf_error)
  {
    case GCONF_ERROR_SUCCESS:
      err = SETTING_ERROR_SUCCESS;
      break;
    case GCONF_ERROR_NO_PERMISSION:
      err = SETTING_ERROR_NO_PERM;
      break;
    case GCONF_ERROR_BAD_KEY:
      err = SETTING_ERROR_BAD_KEY;
      break;
    case GCONF_ERROR_PARSE_ERROR:
      err = SETTING_ERROR_PARSE_ERROR;
      break;
    case GCONF_ERROR_TYPE_MISMATCH:
      err = SETTING_ERROR_TYPE_MISMATCH;
      break;
    default:
      err = SETTING_ERROR_ENTRY_NOT_EXIST;
      break;
  }
  return err;
}

static type_t type_mapping (GConfValueType gconf_type)
{ 
  type_t type;	
  switch (gconf_type)
  {
  	case GCONF_VALUE_STRING:
  	  type = SETTING_TYPE_STRING;
  	  break;
  	case GCONF_VALUE_BOOL:
  	  type = SETTING_TYPE_BOOLEAN;
  	  break;
  	case GCONF_VALUE_INT:
  	  type = SETTING_TYPE_INT;
  	  break;
  	default:
  	  type = SETTING_TYPE_INVALID;
  	  break;  	        
  }
  return type;
}

static void     gconf_client_notify_func    (GConfClient *client,
                                             guint cnxn_id,
                                             GConfEntry *entry,
                                             gpointer user_data)
{
  int cnid = cnxn_id;	
  int cursor = cnid_array_lookup_cursor (cnid);
  g_return_if_fail (cursor >= 0);
  Notify *notify = (Notify*)g_hash_table_lookup (hash_cnid_notify, &cnid_array [cursor]);  
  int setting = GPOINTER_TO_INT (g_hash_table_lookup (hash_cnid_setting, &cnid_array [cursor]));  
  gchar *key = entry->key;
  printf ("%s:%s has its value changed\n", __FUNCTION__, key);
  type_t type = type_mapping (entry->value->type);
  action_t action = SETTING_ACTION_CHANGE;
  g_return_if_fail (notify != NULL); 
  (*notify->func) (setting, cnxn_id, key, type, action, user_data);
}

 /**
  * setting_open:
  * 
  * Get the handle to access the setting system, the returned value should be used in subsequent setting function calls.
  * 
  * Return value: a handle to the setting backend
  **/ 
int setting_open (void)
{
  
  if (!hash_gconfclient)
    hash_gconfclient = g_hash_table_new_full(g_int_hash, g_int_equal, NULL, g_object_unref);
  
  if (hash_gconfclient)
  {
    setting ++;
    GConfClient* gconfclient = gconf_client_get_default ();
    if (gconfclient)
    {
      g_hash_table_insert (hash_gconfclient, &setting, (gpointer) gconfclient);
      return setting;
    }
  }
  
  return -1;

}

/**
 * setting_close:
 * @setting: the setting handle returned from setting_open()
 *
 * Terminates the connection with the setting backend.
 * 
 **/

void setting_close (int setting)
{
  g_hash_table_remove (hash_gconfclient, &setting);
}


/**
 * setting_notify_add:  
 * @setting: the Setting handle.
 * @key: includes the key(s) in which your application is interested.
 * @func: the function to be called to handle value changes of the keys in branch.
 * @user_data: user provided data.
 * @id: conncetion ID for later removing of the notify.
 * 
 * Register a callback for the value change of certain entriess. When changes happen to the entry(ies) represented (contained) by #key, 
 * #func will be invoked. Each call to setting_notify_add will return an integer which represents the connection ID for that notification.
 * This connection ID can be used later by setting_notify_remove.
 * 
 * Return value: error.
 **/
error_t setting_notify_add (int setting, const char *key, SettingNotifyFunc func, void *user_data, int *id)
{  
  if (!hash_cnid_notify)
    hash_cnid_notify = g_hash_table_new_full(g_int_hash, g_int_equal, NULL, notify_destroy);

  if (!hash_cnid_setting)
    hash_cnid_setting = g_hash_table_new_full(g_int_hash, g_int_equal, NULL, NULL);
    
  GConfClient *client = (GConfClient*) g_hash_table_lookup (hash_gconfclient, &setting);
  if (client)
  {
    GError *error = NULL;
    int cnid = gconf_client_notify_add (client, key, gconf_client_notify_func, (gpointer) user_data, NULL, &error);    
       
    if (error == NULL)
    {
      Notify* notify = notify_new (func, NULL, NULL);  
      cnid_array [cnid_array_cursor] = cnid;    
      g_hash_table_insert (hash_cnid_notify, &cnid_array [cnid_array_cursor], (gpointer) notify);
      g_hash_table_insert (hash_cnid_setting, &cnid_array [cnid_array_cursor], GINT_TO_POINTER (setting));
      cnid_array_cursor ++;
      cnid_array_cursor = cnid_array_cursor % MAX_CNXN;
      *id = cnid;      
    }
    error_t ret = error_mapping (error);
    if (error)
      g_error_free (error);    
    gconf_client_add_dir (client, key, GCONF_CLIENT_PRELOAD_ONELEVEL, NULL);
    
    return ret; 
  }
  else
  {
    return SETTING_ERROR_BACKEND_LOST;
  }  
}
/**
 * setting_notify_remove: 
 * @setting: the Setting handle.
 * @id: returned from setting_notify_add.
 * 
 * Revoke the callback added by setting_notify_add.
 **/
void setting_notify_remove (int setting, int id)
{
  GConfClient *client = (GConfClient*) g_hash_table_lookup (hash_gconfclient, &setting);
  if (client)
    gconf_client_notify_remove (client, id);
}

/**
 * setting_set_int:
 * @setting: the Setting handle.
 * @key: the key representing an entry.
 * @value: the new value of this entry.
 * 
 * Set the value of an entry whose type is SETTING_TYPE_INT.
 * 
 * Return value: error number. 
 **/ 
error_t setting_set_int (int setting, const char *key, int value)
{
  GConfClient *client = (GConfClient*) g_hash_table_lookup (hash_gconfclient, &setting);
  if (client)
  {
  	GError *error = NULL;
    gconf_client_set_int (client, key, value, &error);
    error_t ret = error_mapping (error);
  	if (error)
  	  g_error_free (error);
  	return ret; 
  }
  else
  {
  	return SETTING_ERROR_BACKEND_LOST;
  }
  
}
/**
 * setting_set_boolean:
 * @setting: the Setting handle.
 * @key: the key representing an entry.
 * @value: the new value of this entry.
 * 
 * Set the value of an entry whose type is SETTING_TYPE_BOOLEAN.
 * 
 * Return value: error number. 
 **/ 
error_t setting_set_boolean (int setting, const char *key, boolean value)
{
  GConfClient *client = (GConfClient*) g_hash_table_lookup (hash_gconfclient, &setting);
  if (client)
  {
  	GError *error = NULL;
    gconf_client_set_bool (client, key, value, &error);
    error_t ret = error_mapping (error);
  	if (error)
  	  g_error_free (error);
  	return ret;   
  }
  else
  {
    return SETTING_ERROR_BACKEND_LOST;
  }
}

/**
 * setting_set_string:
 * @setting: the Setting handle.
 * @key: the key representing an entry.
 * @value: the new value of this entry.
 * 
 * Set the value of an entry whose type is SETTING_TYPE_STRING.
 * 
 * Return value: error number. 
 **/ 
error_t setting_set_string (int setting, const char *key, const char * value)
{
  GConfClient *client = (GConfClient*) g_hash_table_lookup (hash_gconfclient, &setting);
  if (client)
  {
  	GError *error = NULL;
    gconf_client_set_string (client, key, value, &error);
    error_t ret = error_mapping (error);
  	if (error)
  	  g_error_free (error);
  	return ret; 
  }
  else
  {
    return SETTING_ERROR_BACKEND_LOST;
  }
}

/**
 * setting_get_int:
 * @setting: the Setting handle.
 * @key: the key representing an entry.
 * @value: the return location for the value of this entry.
 * 
 * Get the value of an entry whose type is SETTING_TYPE_INT.
 * 
 * Return value: error number. 
 **/ 
error_t setting_get_int (int setting, const char * key, int *value)
{  
  GConfClient *client = (GConfClient*) g_hash_table_lookup (hash_gconfclient, &setting);
  if (client)
  {
  	GError *error = NULL;
  	*value = gconf_client_get_int (client, key, &error);
  	error_t ret = error_mapping (error);
  	if (error)
  	  g_error_free (error);
  	return ret;  	
  }
  else
  {
    return SETTING_ERROR_BACKEND_LOST;
  }
}

/**
 * setting_get_boolean:
 * @setting: the Setting handle.
 * @key: the key representing an entry.
 * @value: the return location for the value of this entry.
 * 
 * Get the value of an entry whose type is SETTING_TYPE_BOOLEAN.
 * 
 * Return value: error number. 
 **/ 
error_t setting_get_boolean (int setting, const char * key, boolean *value)
{    
  GConfClient *client = (GConfClient*) g_hash_table_lookup (hash_gconfclient, &setting);
  if (client)
  {
  	GError *error = NULL;
  	*value = gconf_client_get_bool (client, key, &error);
  	error_t ret = error_mapping (error);
  	if (error)
  	  g_error_free (error);
  	return ret;  	
  }
  else
  {
    return SETTING_ERROR_BACKEND_LOST;
  }
}

/**
 * setting_get_string:
 * @setting: the Setting handle.
 * @key: the key representing an entry.
 * @value: the return location for the value of this entry, on error, NULL is returned.
 * 
 * Get the value of an entry whose type is SETTING_TYPE_STRING. 
 * The returned string should be freed when no longer needed.
 * 
 * Return value: error number. 
 **/ 
error_t setting_get_string (int setting, const char * key, char ** value)
{     	
  GConfClient *client = (GConfClient*) g_hash_table_lookup (hash_gconfclient, &setting);
  if (client)
  {
  	GError *error = NULL;
  	*value = gconf_client_get_string (client, key, &error);
  	error_t ret = error_mapping (error);
  	if (error)
  	  g_error_free (error);
  	return ret;  	
  }
  else
  {
    return SETTING_ERROR_BACKEND_LOST;
  }
}

/**
 * setting_entry_exist:
 * @setting: the Setting handle.
 * @key: the key representing an entry.
 * 
 * Check whether an entry exists.
 * 
 * Return value: %TRUE if entry exists
 * 
 **/ 
boolean setting_entry_exist (int setting, const char * key)
{
  GConfClient *client = (GConfClient*) g_hash_table_lookup (hash_gconfclient, &setting);
  if (client)
  {
  	return gconf_client_dir_exists (client, key, NULL);
  }	
  else
    return FALSE;
}

/**
 * setting_entry_get_type:
 * @setting: the Setting handle.
 * @key: the key representing an entry.
 * @type: the return location for the type of this entry 
 * 
 * Get the type of an entry
 * 
 * Return value: error number
 * 
 **/ 
error_t setting_entry_get_type (int setting, const char *key, type_t *type)
{
  GConfClient *client = (GConfClient*) g_hash_table_lookup (hash_gconfclient, &setting);
  if (client)
  {  
  	GError *error = NULL;	
  	GConfValue *confvalue = gconf_client_get (client, key, &error);
  	if (confvalue)
  	{
  	  *type = type_mapping (confvalue->type);	
  	}
  	else
  	{ 
  	  *type = SETTING_TYPE_INVALID;
  	}
  	error_t ret = error_mapping (error);
  	if (error)
  	  g_error_free (error);
  	return ret; 
  }
  else
  {
  	*type = SETTING_TYPE_INVALID;
    return SETTING_ERROR_BACKEND_LOST;
  }
}

